# -*- coding: utf-8 -*-
# Standard library imports
from json import dumps as json_dump

try:
    from urllib2 import urlopen, HTTPError, URLError
    from httplib import HTTPException
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError
    from http.client import HTTPException

try:
    import certifi
except ImportError:
    from .vendor import certifi

import ssl


def get(url):
    '''Make a get request.'''

    context = ssl.create_default_context(cafile=certifi.where())
    response = urlopen(url, context=context)
    return response


def json(response):
    '''Get dict from json response.'''

    return json_dump(response.read())



#! /opt/local/bin/python
 
# XML-RPC CLIENT LIBRARY
# $Id: xmlrpclib.py 41594 2005-12-04 19:11:17Z andrew.kuchling $
#
# an XML-RPC client interface for Python.
#
# the marshalling and response parser code can also be used to
# implement XML-RPC servers.
#
# Notes:
# this version is designed to work with Python 2.1 or newer.
#
# History:
# 1999-01-14 fl  Created
# 1999-01-15 fl  Changed dateTime to use localtime
# 1999-01-16 fl  Added Binary/base64 element, default to RPC2 service
# 1999-01-19 fl  Fixed array data element (from Skip Montanaro)
# 1999-01-21 fl  Fixed dateTime constructor, etc.
# 1999-02-02 fl  Added fault handling, handle empty sequences, etc.
# 1999-02-10 fl  Fixed problem with empty responses (from Skip Montanaro)
# 1999-06-20 fl  Speed improvements, pluggable parsers/transports (0.9.8)
# 2000-11-28 fl  Changed boolean to check the truth value of its argument
# 2001-02-24 fl  Added encoding/Unicode/SafeTransport patches
# 2001-02-26 fl  Added compare support to wrappers (0.9.9/1.0b1)
# 2001-03-28 fl  Make sure response tuple is a singleton
# 2001-03-29 fl  Don't require empty params element (from Nicholas Riley)
# 2001-06-10 fl  Folded in _xmlrpclib accelerator support (1.0b2)
# 2001-08-20 fl  Base xmlrpclib.Error on built-in Exception (from Paul Prescod)
# 2001-09-03 fl  Allow Transport subclass to override getparser
# 2001-09-10 fl  Lazy import of urllib, cgi, xmllib (20x import speedup)
# 2001-10-01 fl  Remove containers from memo cache when done with them
# 2001-10-01 fl  Use faster escape method (80% dumps speedup)
# 2001-10-02 fl  More dumps microtuning
# 2001-10-04 fl  Make sure import expat gets a parser (from Guido van Rossum)
# 2001-10-10 sm  Allow long ints to be passed as ints if they don't overflow
# 2001-10-17 sm  Test for int and long overflow (allows use on 64-bit systems)
# 2001-11-12 fl  Use repr() to marshal doubles (from Paul Felix)
# 2002-03-17 fl  Avoid buffered read when possible (from James Rucker)
# 2002-04-07 fl  Added pythondoc comments
# 2002-04-16 fl  Added __str__ methods to datetime/binary wrappers
# 2002-05-15 fl  Added error constants (from Andrew Kuchling)
# 2002-06-27 fl  Merged with Python CVS version
# 2002-10-22 fl  Added basic authentication (based on code from Phillip Eby)
# 2003-01-22 sm  Add support for the bool type
# 2003-02-27 gvr Remove apply calls
# 2003-04-24 sm  Use cStringIO if available
# 2003-04-25 ak  Add support for nil
# 2003-06-15 gn  Add support for time.struct_time
# 2003-07-12 gp  Correct marshalling of Faults
# 2003-10-31 mvl Add multicall support
# 2004-08-20 mvl Bump minimum supported Python version to 2.1
#
# Copyright (c) 1999-2002 by Secret Labs AB.
# Copyright (c) 1999-2002 by Fredrik Lundh.
#
# info@pythonware.com
# http://www.pythonware.com
#
# --------------------------------------------------------------------
# The XML-RPC client interface is
#
# Copyright (c) 1999-2002 by Secret Labs AB
# Copyright (c) 1999-2002 by Fredrik Lundh
#
# By obtaining, using, and/or copying this software and/or its
# associated documentation, you agree that you have read, understood,
# and will comply with the following terms and conditions:
#
# Permission to use, copy, modify, and distribute this software and
# its associated documentation for any purpose and without fee is
# hereby granted, provided that the above copyright notice appears in
# all copies, and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of
# Secret Labs AB or the author not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# SECRET LABS AB AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANT-
# ABILITY AND FITNESS.  IN NO EVENT SHALL SECRET LABS AB OR THE AUTHOR
# BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
# --------------------------------------------------------------------

#
# things to look into some day:

# TODO: sort out True/False/boolean issues for Python 2.3

"""
An XML-RPC client interface for Python.

The marshalling and response parser code can also be used to
implement XML-RPC servers.

Exported exceptions:
  
  Error          Base class for client errors
  ProtocolError  Indicates an HTTP protocol error
  ResponseError  Indicates a broken response package
  Fault          Indicates an XML-RPC fault package

Exported classes:
  
  ServerProxy    Represents a logical connection to an XML-RPC server
  
  MultiCall      Executor of boxcared xmlrpc requests
  Boolean        boolean wrapper to generate a "boolean" XML-RPC value
  DateTime       dateTime wrapper for an ISO 8601 string or time tuple or
                 localtime integer value to generate a "dateTime.iso8601"
                 XML-RPC value
  Binary         binary data wrapper
  
  SlowParser     Slow but safe standard parser (based on xmllib)
  Marshaller     Generate an XML-RPC params chunk from a Python data structure
  Unmarshaller   Unmarshal an XML-RPC response from incoming XML event message
  Transport      Handles an HTTP transaction to an XML-RPC server
  SafeTransport  Handles an HTTPS transaction to an XML-RPC server

Exported constants:
  
  True
  False

Exported functions:
  
  boolean        Convert any Python value to an XML-RPC boolean
  getparser      Create instance of the fastest available parser & attach
                 to an unmarshalling object
  dumps          Convert an argument tuple or a Fault instance to an XML-RPC
                 request (or response, if the methodresponse option is used).
  loads          Convert an XML-RPC packet to unmarshalled data plus a method
                 name (None if not present).
"""

import re, string, time, operator

from types import *
import socket
import errno
import httplib

# --------------------------------------------------------------------
# Internal stuff

try:
    unicode
except NameError:
    unicode = None # unicode support not available

try:
    import datetime
    #import sg_timezone
except ImportError:
    datetime = None

try:
    _bool_is_builtin = False.__class__.__name__ == "bool"
except NameError:
    _bool_is_builtin = 0

def _decode(data, encoding, is8bit=re.compile("[\x80-\xff]").search):
    # decode non-ascii string (if possible)
    if unicode and encoding and is8bit(data):
        data = unicode(data, encoding)
    return data

def escape(s, replace=string.replace):
    s = replace(s, "&", "&amp;")
    s = replace(s, "<", "&lt;")
    return replace(s, ">", "&gt;",)

if unicode:
    def _stringify(string):
        # convert to 7-bit ascii if possible
        try:
            return string.encode("ascii")
        except UnicodeError:
            return string
else:
    def _stringify(string):
        return string

#__version__ = "1.0.1"

# xmlrpc integer limits
MAXINT =  2L**31-1
MININT = -2L**31

# --------------------------------------------------------------------
# Error constants (from Dan Libby's specification at
# http://xmlrpc-epi.sourceforge.net/specs/rfc.fault_codes.php)

# Ranges of errors
PARSE_ERROR       = -32700
SERVER_ERROR      = -32600
APPLICATION_ERROR = -32500
SYSTEM_ERROR      = -32400
TRANSPORT_ERROR   = -32300

# Specific errors
NOT_WELLFORMED_ERROR  = -32700
UNSUPPORTED_ENCODING  = -32701
INVALID_ENCODING_CHAR = -32702
INVALID_XMLRPC        = -32600
METHOD_NOT_FOUND      = -32601
INVALID_METHOD_PARAMS = -32602
INTERNAL_ERROR        = -32603

# --------------------------------------------------------------------
# Exceptions

##
# Base class for all kinds of client-side errors.

class Error(Exception):
    """Base class for client errors."""
    def __str__(self):
        return repr(self)

##
# Indicates an HTTP-level protocol error.  This is raised by the HTTP
# transport layer, if the server returns an error code other than 200
# (OK).
#
# @param url The target URL.
# @param errcode The HTTP error code.
# @param errmsg The HTTP error message.
# @param headers The HTTP header dictionary.

class ProtocolError(Error):
    """Indicates an HTTP protocol error."""
    def __init__(self, url, errcode, errmsg, headers):
        Error.__init__(self)
        self.url = url
        self.errcode = errcode
        self.errmsg = errmsg
        self.headers = headers
    def __repr__(self):
        return (
            "<ProtocolError for %s: %s %s>" %
            (self.url, self.errcode, self.errmsg)
            )

##
# Indicates a broken XML-RPC response package.  This exception is
# raised by the unmarshalling layer, if the XML-RPC response is
# malformed.

class ResponseError(Error):
    """Indicates a broken response package."""
    pass

##
# Indicates an XML-RPC fault response package.  This exception is
# raised by the unmarshalling layer, if the XML-RPC response contains
# a fault string.  This exception can also used as a class, to
# generate a fault XML-RPC message.
#
# @param faultCode The XML-RPC fault code.
# @param faultString The XML-RPC fault string.

class Fault(Error):
    """Indicates an XML-RPC fault package."""
    def __init__(self, faultCode, faultString, **extra):
        Error.__init__(self)
        self.faultCode = faultCode
        self.faultString = faultString
    def __repr__(self):
        return (
            "<Fault %s: %s>" %
            (self.faultCode, repr(self.faultString))
            )

# --------------------------------------------------------------------
# Special values

##
# Wrapper for XML-RPC boolean values.  Use the xmlrpclib.True and
# xmlrpclib.False constants, or the xmlrpclib.boolean() function, to
# generate boolean XML-RPC values.
#
# @param value A boolean value.  Any true value is interpreted as True,
#              all other values are interpreted as False.

if _bool_is_builtin:
    boolean = Boolean = bool
    # to avoid breaking code which references xmlrpclib.{True,False}
    True, False = True, False
else:
    class Boolean:
        """Boolean-value wrapper.
        
        Use True or False to generate a "boolean" XML-RPC value.
        """
        
        def __init__(self, value = 0):
            self.value = operator.truth(value)
        
        def encode(self, out):
            out.write("<value><boolean>%d</boolean></value>\n" % self.value)
        
        def __cmp__(self, other):
            if isinstance(other, Boolean):
                other = other.value
            return cmp(self.value, other)
        
        def __repr__(self):
            if self.value:
                return "<Boolean True at %x>" % id(self)
            else:
                return "<Boolean False at %x>" % id(self)
        
        def __int__(self):
            return self.value
        
        def __nonzero__(self):
            return self.value
    
    True, False = Boolean(1), Boolean(0)
    
    ##
    # Map true or false value to XML-RPC boolean values.
    #
    # @def boolean(value)
    # @param value A boolean value.  Any true value is mapped to True,
    #              all other values are mapped to False.
    # @return xmlrpclib.True or xmlrpclib.False.
    # @see Boolean
    # @see True
    # @see False
    
    def boolean(value, _truefalse=(False, True)):
        """Convert any Python value to XML-RPC 'boolean'."""
        return _truefalse[operator.truth(value)]





# -*- coding: utf-8 -*-
'''isort:skip_file'''
__title__ = 'cpenv'
__version__ = '0.5.14'
__author__ = 'Dan Bradham'
__email__ = 'danielbradham@gmail.com'
__url__ = 'http://github.com/cpenv/cpenv'
__license__ = 'MIT'
__description__ = 'Cross-platform Python environment management.'


# Local imports
from .api import *
from .module import *
from .repos import *
from .reporter import *
from .resolver import *
from .versions import *
from . import vendor
from . import paths
from . import compat
from . import mappings


# Initialize cpenv
from .api import _init
_init()



# -*- coding: utf-8 -*-
import sys
import contextlib

__all__ = [
    'set_reporter',
    'get_reporter',
    'Reporter',
]
this = sys.modules[__name__]
this._reporter = None


def set_reporter(reporter, *args, **kwargs):
    if isinstance(reporter, Reporter):
        this._reporter = reporter
    else:
        this._reporter = reporter(*args, **kwargs)


def get_reporter():
    if this._reporter is None:
        this._reporter = Reporter()
    return this._reporter


class ProgressBar(object):

    def __init__(self, reporter, label, max_size, data):
        self.reporter = reporter
        self.label = label
        self.max_size = max_size
        self.data = data

    def start(self):
        self.reporter.start_progress(self.label, self.max_size, self.data)

    def update(self, chunk_size=0, data=None):
        if data is not None:
            self.data = data
        self.reporter.update_progress(self.label, chunk_size, self.data)

    def end(self):
        self.reporter.end_progress(self.label, self.data)


class Reporter(object):

    ProgressBar = ProgressBar

    def start_resolve(self, requirements):
        '''Called when a Resolver.resolve is called with some requirements.'''

    def find_requirement(self, requirement):
        '''Called just before attempting to resolve a requirement.'''

    def resolve_requirement(self, requirement, module_spec):
        '''Called when a a requirement is resolved.'''

    def end_resolve(self, resolved, unresolved):
        '''Called when Resolver.resolve is done.'''

    def start_localize(self, module_specs):
        '''Called when Localizer.localize is called with a list of specs.'''

    def localize_module(self, module_spec, module):
        '''Called when a module is localized.'''

    def end_localize(self, localized):
        '''Called when Localizer.localize is done.'''

    def start_progress(self, label, max_size, data):
        '''Called when a download is started.'''

    def update_progress(self, label, chunk_size, data):
        '''Called each time a chunk is downloaded.'''

    def end_progress(self, label, data):
        '''Called when a download is finished.'''

    @contextlib.contextmanager
    def progress_bar(self, label, max_size, data=None):
        bar = self.ProgressBar(self, label, max_size, data)
        try:
            bar.start()
            yield bar
        finally:
            bar.end()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import fnmatch
import os
import subprocess

# Local imports
from . import compat


def binpath(*paths):
    '''Like os.path.join but acts relative to this packages bin path.'''

    package_root = os.path.dirname(__file__)
    return os.path.normpath(os.path.join(package_root, 'bin', *paths))


def run(*args, **kwargs):
    '''Returns True if successful, False if failure'''

    kwargs.setdefault('env', os.environ)
    kwargs.setdefault('shell', True)

    try:
        subprocess.check_call(' '.join(args), **kwargs)
        return True
    except subprocess.CalledProcessError:
        print('Error running: {}'.format(args))
        return False


def get_shell():
    '''Get an approriate shell to use for a subshell.

    1. Use value of CPENV_SUBSHELL env var
    2. Attempt to walk up process tree using psutil
    3. Use a default shell (May not match with parent shell)

    Default Shells:
        win: powershell.exe
        mac: bash
        linux: bash
    '''

    # Get preferred shell from
    preferred_shell = os.getenv('CPENV_SUBSHELL')
    if preferred_shell:
        return preferred_shell

    # Attempt to find parent shell using psutil
    try:
        import psutil
        supported_shells = [
            'powershell.exe',
            'cmd.exe',
            'powershell',
            'cmd',
            'bash',
            'sh',
            'fish',
            'xonsh',
            'zsh',
            'csh',
        ]

        if compat.platform == 'win':
            py_proc = 'py*.exe'
        else:
            py_proc = 'python*'

        proc = psutil.Process()
        while True:
            exe = proc.exe()
            exe_name = os.path.basename(exe)
            if fnmatch.fnmatch(exe_name, py_proc):
                proc = proc.parent()
                continue
            if exe_name in supported_shells:
                return exe
            break
    except ImportError:
        pass

    # Fallback to a default shell
    return {
        'win': 'powershell.exe',
        'linux': 'bash',
        'mac': 'bash',
    }[compat.platform]


def get_prompt(shell, prefix, colored=True):
    '''Generate a prompt with a given prefix

    linux/osx: [prefix] user@host cwd $
          win: [prefix] cwd:
    '''

    if shell.endswith('cmd.exe'):
        return '{} $P$G'.format(prefix)

    if shell.endswith('powershell.exe'):
        return "'{} PS ' + $(get-location) + '> '".format(prefix)

    if shell.endswith('bash'):
        if colored:
            return (
                '{} '  # White prefix
                '\\[\\033[01;32m\\]\\u@\\h\\[\\033[00m\\] '  # Green user@host
                '\\[\\033[01;34m\\]\\w $ \\[\\033[00m\\]'  # Blue cwd $
            ).format(prefix)
        return '{} \\u@\\h \\w $ '.format(prefix)


def get_subshell_command(prefix):
    '''Return a command to launch a subshell'''

    shell = get_shell()
    prompt = get_prompt(shell, prefix)

    if shell.endswith('cmd.exe'):
        os.environ['PROMPT'] = prompt
        return [shell, '/K']

    if shell.endswith('powershell.exe'):
        return [
            shell,
            '-NoExit',
            '-Command',
            "function Prompt {%s}" % prompt
        ]

    if shell.endswith('bash'):
        return [shell, binpath('subshell.sh')]

    return [shell]


def launch(prefix='[*]'):
    '''Launch a subshell'''

    shell_cmd = get_subshell_command(prefix)
    subprocess.call(shell_cmd, env=dict(os.environ))



"""
 -----------------------------------------------------------------------------
 Copyright (c) 2009-2017, Shotgun Software Inc

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

  - Neither the name of the Shotgun Software Inc nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

-----------------------------------------------------------------------------
"""

from ..six.moves import cPickle as pickle
import os

from .errors import MockgunError


class SchemaFactory(object):
    """
    Allows to instantiate a pickled schema.
    """

    _schema_entity_cache = None
    _schema_entity_cache_path = None
    _schema_cache = None
    _schema_cache_path = None

    @classmethod
    def get_schemas(cls, schema_path, schema_entity_path):
        """
        Retrieves the schemas from disk.

        :param str schema_path: Path to the schema.
        :param str schema_entity_path: Path to the entities schema.

        :returns: Pair of dictionaries holding the schema and entities schema.
        :rtype: tuple
        """
        if not os.path.exists(schema_path):
            raise MockgunError("Cannot locate Mockgun schema file '%s'!" % schema_path)

        if not os.path.exists(schema_entity_path):
            raise MockgunError("Cannot locate Mockgun schema file '%s'!" % schema_entity_path)

        # Poor man's attempt at a cache. All of our use cases deal with a single pair of files
        # for the duration of the unit tests, so keep a cache for both inputs. We don't want
        # to deal with ever growing caches anyway. Just having this simple cache has shown
        # speed increases of up to 500% for Toolkit unit tests alone.

        if schema_path != cls._schema_cache_path:
            cls._schema_cache = cls._read_file(schema_path)
            cls._schema_cache_path = schema_path

        if schema_entity_path != cls._schema_entity_cache_path:
            cls._schema_entity_cache = cls._read_file(schema_entity_path)
            cls._schema_entity_cache_path = schema_entity_path

        return cls._schema_cache, cls._schema_entity_cache

    @classmethod
    def _read_file(cls, path):
        fh = open(path, "rb")
        try:
            return pickle.load(fh)
        finally:
            fh.close()


# Highest protocol that Python 2.4 supports, which is the earliest version of Python we support.
# Actually, this is the same version that Python 2.7 supports at the moment!
_HIGHEST_24_PICKLE_PROTOCOL = 2


# ----------------------------------------------------------------------------
# Utility methods
def generate_schema(shotgun, schema_file_path, schema_entity_file_path):
    """
    Helper method for mockgun.
    Generates the schema files needed by the mocker by connecting to a real shotgun
    and downloading the schema information for that site. Once the generated schema
    files are being passed to mockgun, it will mimic the site's schema structure.

    :param sg_url: Shotgun site url
    :param sg_script: Script name to connect with
    :param sg_key: Script key to connect with
    :param schema_file_path: Path where to write the main schema file to
    :param schema_entity_file_path: Path where to write the entity schema file to
    """

    schema = shotgun.schema_read()
    fh = open(schema_file_path, "wb")
    try:
        pickle.dump(schema, fh, protocol=_HIGHEST_24_PICKLE_PROTOCOL)
    finally:
        fh.close()

    schema_entity = shotgun.schema_entity_read()
    fh = open(schema_entity_file_path, "wb")
    try:
        pickle.dump(schema_entity, fh, protocol=_HIGHEST_24_PICKLE_PROTOCOL)
    finally:
        fh.close()



# -*- coding: utf-8 -*-


class Repo(object):
    '''Base class for all Repos.

    A Repo is a source of modules. They can be local or remote so long as they
    provide this interface.
    '''

    type_name = 'repo'

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        args = []
        for attr in ['name', 'path']:
            if hasattr(self, attr):
                args.append('{}={!r}'.format(attr, getattr(self, attr)))
        return '<{}>({})'.format(type(self).__name__, ', '.join(args))

    def clear_cache(self):
        '''Subclasses that implement caching should implement this method
        so that cpenv can clear the cache when necessary. Like after copying
        or localizing modules.
        '''
        return NotImplemented

    def find(self, requirement):
        '''Given a requirement, return a list of ModuleSpecs that match.

        Should be ordered from best to worst match.
        '''

        return NotImplemented

    def list(self):
        '''Return a list of ModuleSpecs in this Repo.'''

        return NotImplemented

    def download(self, module_spec, where, overwrite=False):
        '''Given a module_spec and output path, download it from this Repo.

        Return a newly downloaded Module.
        '''

        return NotImplemented

    def upload(self, module, overwrite=False):
        '''Given a module, upload it to this Repo.

        Return a ModuleSpec.
        '''

        return NotImplemented

    def remove(self, module_spec):
        '''Given a module_spec, remove it from this repo.'''

        return NotImplemented

    def get_data(self, module_spec):
        '''Given a module_spec, return a Module's config(module.yml) dict.

        If possible this should be implemented for all Repos, however, it
        is not used during module resolution. It is only used to provide
        a richer view of Modules in Repos by UI implementations.
        '''

        return {}

    def get_size(self, module_spec):
        '''Given a module_spec, return the size of a module in the repo.

        For remote repositories this may mean reading a field in a databse.
        For the LocalRepo it means getting the combined size of the module
        folder's contents.

        Repos should implement this method to full support progress bars
        while publishing and localizing modules.

        Return -1 when size if the size is unavailable.
        '''
        return -1

    def get_thumbnail(self, module_spec):
        '''Given a module_spec, return the path to a thumbnail.

        Remote repos should download a thumbnail to the cache/icons:
            cache_path = cpenv.get_cache_path('icons', 'cached-0.1.0.png')

        Return None when no thumbnail is available.
        '''
        return



# -*- coding: utf-8 -*-

# Standard library imports
import os
import sys


py_ver = sys.version_info[0]
is_py2 = py_ver == 2
is_py3 = py_ver == 3

if is_py2:
    string_types = (str, basestring, unicode)
    numeric_types = (int, long, float)

if is_py3:
    string_types = (str, bytes)
    numeric_types = (int, float)


platform = sys.platform.rstrip('1234567890').lower()
if platform == 'darwin':  # Use mac instead of darwin
    platform = 'mac'
os.environ['CPENV_PLATFORM'] = platform


# Leave osx in supported platforms for backwards compatability
supported_platforms = ['win', 'linux', 'mac', 'osx']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import collections
import os
import random
import sys
import tempfile
from string import Template

# Local imports
from . import paths
from .compat import numeric_types, platform, supported_platforms, string_types
from .vendor import yaml


env_value_types = numeric_types + string_types
Item = collections.namedtuple('Item', 'key value')
Op = collections.namedtuple('Op', 'key value op')


class CaseInsensitiveDict(collections.MutableMapping):
    '''A case insensitive dict that expects strings as keys.

    Based on the requests.structures.CaseInsensitiveDict.
    '''

    def __init__(self, *args, **kwargs):
        self._items = dict()
        self.update(*args, **kwargs)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, dict(self.items()))

    def __setitem__(self, key, value):
        self._items[key.lower()] = Item(key, value)

    def __getitem__(self, key):
        return self._items[key.lower()].value

    def __delitem__(self, key):
        del self._items[key.lower()]

    def __iter__(self):
        return (item.key for item in self._items.values())

    def __len__(self):
        return len(self._items)

    def __eq__(self, other):
        other_cmp = self._comparable_mapping(other)
        if other_cmp is None:
            return NotImplemented
        self_cmp = self._comparable_mapping(self)
        return self_cmp == other_cmp

    @classmethod
    def _comparable_mapping(cls, mapping):
        if isinstance(mapping, CaseInsensitiveDict):
            return {item.key.lower(): item.value for item in mapping.values()}
        if isinstance(mapping, collections.Mapping):
            return {key.lower(): value for key, value in mapping.items()}


class EnvironmentDict(CaseInsensitiveDict):
    '''A dict suited to storing and manipulating environment variables.

    Lowercase comparisons are used to ensure unique keys and values.

    The following methods make it easy to build an Environment.
    - set: insert or overwrite the value of a key.
    - unset: discard a key.
    - remove: discard an item or list of items from a key's value.
    - append: add an item or list of items to the tail of a key's value.
    - prepend: add an item or list of items to the head of a key's value.

    Example:
        >>> env = EnvironmentDict()
        >>> env.append('PATH', '/some/path')
        >>> env.prepend('path', '/some/other/path')
        >>> env.remove('PATH', '/some/path')
        >>> assert env['PATH'] == ['/some/other/path']
    '''

    _value_error = 'Expected Union[list, Union[str, int, float]] got %s'

    def __init__(self, *args, **kwargs):
        self._add_condition = kwargs.pop('add_condition', ignore_case)
        super(EnvironmentDict, self).__init__(*args, **kwargs)

    def _get_list(self, key):
        value = self.get(key, None)
        if value is None:
            return []
        if isinstance(value, env_value_types):
            return [value]
        if isinstance(value, collections.Sequence):
            return list(value)
        raise ValueError('%s could not be converted to a list.' % value)

    def _coerce_value(self, value):
        if isinstance(value, env_value_types):
            return str(value)
        elif isinstance(value, collections.Sequence):
            result = []
            for v in value:
                if v in (None, ''):
                    continue
                result.append(str(value))
            return result
        else:
            raise ValueError(self._value_error % type(value))

    def _remove_condition(self, current_value, value):
        return not self._add_condition(current_value, value)

    def unset(self, key, value=None):
        '''Unset a key.'''

        self.pop(key, None)

    def set(self, key, value):
        '''Set a key.'''

        self[key] = self._coerce_value(value)

    def remove(self, key, value, remove_check=None):
        '''Remove a value from a key.'''

        value = self._coerce_value(value)

        if key not in self:
            return

        result = self._get_list(key)
        if isinstance(value, env_value_types):
            if self._remove_condition(result, value):
                result.remove(value)
        elif isinstance(value, collections.Sequence):
            for v in value:
                if self._remove_condition(result, v):
                    result.remove(v)

        if not result:
            self.pop(key, None)
        else:
            self[key] = result

    def prepend(self, key, value, add_check=None):
        '''Prepend a value to a key.

        Sets the value if the key does not exist.
        '''

        result = self._get_list(key)
        value = self._coerce_value(value)

        if isinstance(value, env_value_types):
            if self._add_condition(result, value):
                result.insert(0, value)
        elif isinstance(value, collections.Sequence):
            for v in value:
                if self._add_condition(result, v):
                    result.insert(0, v)

        self[key] = result

    def append(self, key, value, add_check=None):
        '''Append a value to a key.

        Sets the value if the key does not exist.
        '''

        result = self._get_list(key)
        value = self._coerce_value(value)

        if isinstance(value, env_value_types):
            if self._add_condition(result, value):
                result.append(value)
        elif isinstance(value, collections.Sequence):
            for v in value:
                if self._add_condition(result, v):
                    result.append(v)

        self[key] = result


class EnvironmentDictTokenizer(object):
    '''Responsible for converting a dict into a list of Operation tokens that
    can be used to merge one dict with another. Used internally by
    `join_dicts`.

    An Op token will be returned for each of the following keys:
    - set: insert or overwrite the value of a key.
    - unset: discard a key.
    - remove: discard an item or list of items from a key's value.
    - append: add an item or list of items to the tail of a key's value.
    - prepend: add an item or list of items to the head of a key's value.

    You may specify platform specific keys, the value of the current platform
    will be extracted before tokenization.
    - win
    - osx
    - linux

    Example:
        >>> data = {
        ...     'PATH': [
        ...         {'append': '/path/a'},
        ...         {'remove': ['/path/b', '/path/c']}
        ...     ],
        ... }
        [Op('PATH', '/path/a', 'append'), Op('PATH', '/path/b', 'remove')...]
    '''

    operation_names = ['unset', 'set', 'remove', 'append', 'prepend']

    @staticmethod
    def _is_explicit_list_value(value):
        return len(value) and isinstance(value[0], collections.Mapping)

    @classmethod
    def tokenize_mapping(cls, key, value, op, tokens):
        # Handle platform specific keys
        for plat in supported_platforms:
            if plat in value:
                if platform == 'mac':
                    # mac and osx are valid platform keys for macos
                    plat_value = value.get(platform, value.get('osx', None))
                else:
                    plat_value = value.get(platform, None)
                if plat_value:
                    cls.tokenize_value(key, plat_value, op, tokens)
                return

        # Handle explicit operations keys
        for op, op_value in value.items():
            if op not in cls.operation_names:
                raise KeyError('Invalid key in dict value: %s' % op)
            cls.tokenize_value(key, op_value, op, tokens)

    @classmethod
    def tokenize_str(cls, key, value, op, tokens):
        if isinstance(value, bool):
            value = str(int(value))
        elif isinstance(value, env_value_types):
            value = str(value)
        elif value is None:
            # Skip null values
            return
        else:
            raise ValueError('Got invalid value %s for key %s' % (value, key))
        tokens.append(Op(key, value, op))

    @classmethod
    def tokenize_sequence(cls, key, value, op, tokens):
        if not cls._is_explicit_list_value(value) and op == 'prepend':
            value = value[::-1]

        for item in value:
            cls.tokenize_value(key, item, op, tokens)

    @classmethod
    def tokenize_value(cls, key, value, op=None, tokens=None):
        if tokens is None:
            tokens = []
        if op == 'set':
            tokens.append(Op(key, None, 'unset'))
        if isinstance(value, env_value_types):
            op = op or 'set'
            cls.tokenize_str(key, value, op, tokens)
        elif isinstance(value, collections.Mapping):
            op = op or 'set'
            cls.tokenize_mapping(key, value, op, tokens)
        elif isinstance(value, collections.Sequence):
            if op in (None, 'set'):
                op = 'prepend'
            cls.tokenize_sequence(key, value, op, tokens)
        else:
            raise ValueError('Could not tokenize %s' % value)

        return tokens

    @classmethod
    def tokenize(cls, data):
        tokens = []
        for key, value in data.items():
            tokens.extend(cls.tokenize_value(key, value))
        return tokens


def ignore_case(items, value):
    '''Default add_condition for `EnvironmentDict`.

    Returns True if value is not in items.
    '''

    return value.lower() not in [item.lower() for item in items]


def tokenize_dict(data):
    '''Tokenize a dict returning a list of Ops that can be used to
    join the dict with another.
    '''

    return EnvironmentDictTokenizer.tokenize(data)


def join_dicts(*dicts, **kwargs):
    '''Join a bunch of dicts.

    Arguments:
        *dicts: Dictionaries to merge
        add_condition (fn): Used to check if a value should be added to a key
    '''

    env_dict = EnvironmentDict(**kwargs)
    for data in dicts:
        tokens = tokenize_dict(data)
        for token in tokens:
            if token.op == 'unset':
                env_dict.unset(token.key, token.value)
            elif token.op == 'set':
                env_dict.set(token.key, token.value)
            elif token.op == 'remove':
                env_dict.remove(token.key, token.value)
            elif token.op == 'prepend':
                env_dict.prepend(token.key, token.value)
            elif token.op == 'append':
                env_dict.append(token.key, token.value)
    return dict(env_dict)


def env_to_dict(env, pathsep=os.pathsep):
    '''
    Convert a dict containing environment variables into a standard dict.
    Variables containing multiple values will be split into a list based on
    the argument passed to pathsep.

    :param env: Environment dict like dict(os.environ)
    :param pathsep: Path separator used to split variables
    '''

    out_dict = {}

    for k, v in env.items():
        if pathsep in v:
            out_dict[k] = v.split(pathsep)
        else:
            out_dict[k] = v

    return out_dict


def dict_to_env(d, pathsep=os.pathsep):
    '''
    Convert a python dict to a dict containing valid environment variable
    values.

    :param d: Dict to convert to an env dict
    :param pathsep: Path separator used to join lists(default os.pathsep)
    '''

    out_env = {}

    for k, v in d.items():
        if isinstance(v, list):
            out_env[k] = pathsep.join(v)
        elif isinstance(v, string_types):
            out_env[k] = v
        else:
            raise TypeError('{} not a valid env var type'.format(type(v)))

    return out_env


def expand_envvars(env):
    '''
    Expand all environment variables in an environment dict

    :param env: Environment dict
    '''

    out_env = {}

    for k, v in env.items():
        out_env[k] = Template(v).safe_substitute(env)

    # Expand twice to make sure we expand everything we possibly can
    for k, v in out_env.items():
        out_env[k] = Template(v).safe_substitute(out_env)

    return out_env


def get_store_env_tmp():
    '''Returns an unused random filepath.'''

    tempdir = tempfile.gettempdir()
    temp_name = 'envstore{0:0>3d}'
    temp_path = paths.normalize(
        tempdir,
        temp_name.format(random.getrandbits(9))
    )
    if not os.path.exists(temp_path):
        return temp_path
    else:
        return get_store_env_tmp()


def store_env(path=None):
    '''Encode current environment as yaml and store in path or a temporary
    file. Return the path to the stored environment.
    '''

    path = path or get_store_env_tmp()

    env_dict = yaml.safe_dump(dict(os.environ), default_flow_style=False)

    with open(path, 'w') as f:
        f.write(env_dict)

    return path


def restore_env(env_dict):
    '''Set environment variables in the current python process from a dict
    containing envvars and values.'''

    if hasattr(sys, 'real_prefix'):
        sys.prefix = sys.real_prefix
        del(sys.real_prefix)

    replace_osenviron(expand_envvars(dict_to_env(env_dict)))


def restore_env_from_file(env_file):
    '''Restore the current environment from an environment stored in a yaml
    yaml file.

    :param env_file: Path to environment yaml file.
    '''

    with open(env_file, 'r') as f:
        env_dict = yaml.safe_load(f.read())

    restore_env(env_dict)


def set_env(*env_dicts):
    '''Set environment variables in the current python process from a dict
    containing envvars and values.'''

    old_env_dict = env_to_dict(dict(os.environ))
    new_env_dict = join_dicts(old_env_dict, *env_dicts)
    new_env = dict_to_env(new_env_dict)
    replace_osenviron(expand_envvars(new_env))


def set_env_from_file(env_file):
    '''Restore the current environment from an environment stored in a yaml
    yaml file.

    :param env_file: Path to environment yaml file.
    '''

    with open(env_file, 'r') as f:
        env_dict = yaml.safe_load(f.read())

    if 'environment' in env_dict:
        env_dict = env_dict['environment']

    set_env(env_dict)


def replace_osenviron(env_dict):
    for k in os.environ.keys():
        if k not in env_dict:
            del os.environ[k]

    for k, v in env_dict.items():
        os.environ[k] = v



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

# Standard library imports
import os
from types import ModuleType

# Local imports
from . import paths


class HookFinder(object):
    '''Find python hooks by name in the provided path.

    usage::

        >>> hook_finder = HookFinder('~/.cpenv/hooks')
        >>> hook = hook_finder('precreate')
        >>> hook.run(env=VirtualEnvironment('path/to/env'))
    '''

    def __init__(self, *hook_paths):
        self.hook_paths = hook_paths

    def _find_pyfile(self, hook_name):
        for path in self.hook_paths:
            hook_path = paths.normalize(path, hook_name + '.py')
            if os.path.exists(hook_path):
                return hook_path

    def find(self, hook_name):

        hook_path = self._find_pyfile(hook_name)

        if not hook_path:
            return

        try:
            with open(hook_path, 'r') as f:
                code = compile(f.read(), '', 'exec')
        except SyntaxError as e:
            print('SyntaxError compiling hook: {}'.format(e))
            raise

        hook = ModuleType(hook_name)
        hook.__file__ = hook_path

        try:
            exec(code, hook.__dict__)
            return hook
        except Exception:
            print('Error executing hook: {}'.format(hook_path))
            raise

    __call__ = find


def get_global_hook_path():
    '''Returns the global hook path'''

    from .api import get_home_path
    return paths.normalize(get_home_path(), 'hooks')


def run_global_hook(hook_name, *args):
    '''Attempt to run a global hook by name with args'''

    hook_finder = HookFinder(get_global_hook_path())
    hook = hook_finder(hook_name)
    if hook:
        hook.run(*args)



# Copyright (c) 2019 Shotgun Software Inc.
#
# CONFIDENTIAL AND PROPRIETARY
#
# This work is provided "AS IS" and subject to the Shotgun Pipeline Toolkit
# Source Code License included in this distribution package. See LICENSE.
# By accessing, using, copying or modifying this work you indicate your
# agreement to the Shotgun Pipeline Toolkit Source Code License. All rights
# not expressly granted therein are reserved by Shotgun Software Inc.

from .shotgun import (Shotgun, ShotgunError, ShotgunFileDownloadError, # noqa unused imports
                      ShotgunThumbnailNotReady, Fault,
                      AuthenticationFault, MissingTwoFactorAuthenticationFault,
                      UserCredentialsNotAllowedForSSOAuthenticationFault,
                      ProtocolError, ResponseError, Error, __version__)
from .shotgun import SG_TIMEZONE as sg_timezone # noqa unused imports



"""Utilities for certificate management."""

import os

certifi_available = False
certifi_where = None
try:
    from certifi import where as certifi_where
    certifi_available = True
except ImportError:
    pass

custom_ca_locater_available = False
custom_ca_locater_where = None
try:
    from ca_certs_locater import get as custom_ca_locater_where
    custom_ca_locater_available = True
except ImportError:
    pass


BUILTIN_CA_CERTS = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "cacerts.txt"
)


def where():
    env = os.environ.get("HTTPLIB2_CA_CERTS")
    if env is not None:
        if os.path.isfile(env):
            return env
        else:
            raise RuntimeError("Environment variable HTTPLIB2_CA_CERTS not a valid file")
    if custom_ca_locater_available:
        return custom_ca_locater_where()
    if certifi_available:
        return certifi_where()
    return BUILTIN_CA_CERTS


if __name__ == "__main__":
    print(where())



from __future__ import absolute_import

import random

from .cache import Cache


# random.choice cannot be pickled in Python 2.7
def _choice(seq):
    return random.choice(seq)


class RRCache(Cache):
    """Random Replacement (RR) cache implementation."""

    def __init__(self, maxsize, choice=random.choice, getsizeof=None):
        Cache.__init__(self, maxsize, getsizeof)
        # TODO: use None as default, assing to self.choice directly?
        if choice is random.choice:
            self.__choice = _choice
        else:
            self.__choice = choice

    @property
    def choice(self):
        """The `choice` function used by the cache."""
        return self.__choice

    def popitem(self):
        """Remove and return a random `(key, value)` pair."""
        try:
            key = self.__choice(list(self))
        except IndexError:
            raise KeyError('%s is empty' % self.__class__.__name__)
        else:
            return (key, self.pop(key))



# -*- coding: utf-8 -*-
'''A minimal CLI framework.'''
from __future__ import absolute_import, print_function

# Standard library imports
import argparse
import ast
import sys
from textwrap import wrap
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest
try:
    from os import get_terminal_size
except ImportError:
    from .vendor.terminalsize import get_terminal_size

version = sys.version_info[0]
is_py2 = version == 2
is_py3 = version == 3
missing = object()


class CLI(object):

    name = None
    usage = None
    description = None

    def __init__(self, parent=None):
        self.parent = parent
        self._commands = {c.name: c for c in self.commands()}

    @property
    def name(self):
        return self.__class__.__name__.lower()

    @property
    def description(self):
        return self.__doc__

    @property
    def short_description(self):
        return self.__doc__.strip().split('\n')[0].rsplit('.', 1)[0]

    @property
    def fullname(self):
        name = [self.name]
        if self.parent:
            name.insert(0, self.parent.fullname)
        return ' '.join(name)

    def commands(self):
        return []

    def get_command(self, name):
        return self._commands.get(name, None)

    def format_commands(self):
        if not self._commands:
            return ''

        return format_section(
            'commands:',
            [
                (k, v.short_description)
                for k, v in self._commands.items()
            ],
        )

    @property
    def parser(self):
        if hasattr(self, '_parser'):
            return self._parser

        self._parser = argparse.ArgumentParser(
            prog=self.fullname,
            description=self.description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            usage=self.usage,
            add_help=False,
        )
        self._parser.add_argument(
            '-h',
            '--help',
            help='show this help message and exit',
            action='store_true',
            dest='help'
        )
        self._parser._positionals.title = 'required'
        self._parser._optionals.title = 'optional'
        self.setup_parser(self._parser)

        argument_names = [a.dest for a in self._parser._actions]
        if self._commands and 'command' not in argument_names:
            self._parser.add_argument(
                'command',
                help='subcommand to run.',
                action='store',
                nargs='?',
            )
            self._parser.epilog = self.format_commands()

        return self._parser

    def setup_parser(self, parser):
        '''Subclasses implement this method to parse arguments.'''

    def run(self, args):
        '''Subclasses implement this method.'''
        self.parser.print_help()


def echo(message='', end='\n'):
    '''Write a message to stdout.'''

    print(message, end=end)
    if not end:
        sys.stdout.flush()


def prompt(message):
    '''Prompt a user for input'''

    if is_py2:
        return raw_input(message)

    return input(message)


def format_section(header, rows):

    width = max(len(max(rows, key=lambda x: len(x[0]))), 8) + 2

    lines = [header]
    for key, value in rows:

        # Handle lists - one value per line
        additional_lines = []
        if isinstance(value, (list, tuple)):
            if len(value) == 1:
                value = value[0]
            elif len(value) > 1:
                additional_lines = value[1:]
                value = value[0]

        line = wrap(
            '{1:<{0}}  {2}'.format(width, key, value),
            initial_indent='  ',
            subsequent_indent=' ' * (width + 4),
        )
        lines.append('\n'.join(line))

        for value in additional_lines:
            line = wrap(
                '{1:<{0}}  {2}'.format(width, '', value),
                initial_indent='  ',
                subsequent_indent=' ' * (width + 4),
            )
            lines.append('\n'.join(line))

    return '\n'.join(lines)


def columns_to_rows(items, columns, width):

    row_count = int(len(items) / columns + 1)

    # Build columns
    columns = []
    for i in range(0, len(items), row_count):
        column = [elide(item, width) for item in items[i:i + row_count]]
        if len(column) < row_count:
            column += [''] * (row_count - len(column))
        columns.append(column)

    # Transpose for rows
    rows = [list(row) for row in zip_longest(*columns)]
    return rows


def format_columns(header, items, indent='    '):

    size = get_terminal_size()
    width = max(len(max(items, key=len)), 18) + 4
    columns = int((size.columns) / (width + 4))

    lines = [header]
    for row in columns_to_rows(items, columns, width):
        if len(row) < columns:
            row += [''] * (columns - len(row))
        lines.append((
            indent + '{:<{width}} ' * columns
        ).format(*row, width=width))

    return '\n'.join(lines)


def elide(text, width):
    if len(text) > width:
        return text[:width - 3] + '...'
    return text


class NameToString(ast.NodeTransformer):
    '''Visits Name nodes and transforms them to Str nodes'''

    def visit_Name(self, node):
        if node.id in ('True', 'False', 'None'):
            return node
        return ast.copy_location(ast.Str(node.id), node)


def safe_eval(string):
    '''Safely evaluates "string", converting it to a valid python object.
    This is the same technique used by Python Fire to evaluate arguments. This
    method is intended to be used for cli argument conversion and therefore
    converts all ast.Name nodes to ast.Str nodes to reduce the need for
    nested quotes. Note in the *Conversions* section, "{a: 1.0}" and
    "{'a': 1.0}" both evaluate to {'a': 1.0}.
    Arguments:
        string (str): String to convert to a python object
    Conversions:
        "1"            =>  1
        "1.0"          =>  1.0
        "(1,)"         =>  (1,)
        "string"       =>  'string'
        "{a: 1.0}"     =>  {'a': 1.0}
        "{'a': 1.0}"   =>  {'a': 1.0}
        "[a, 1, 2.0]"  =>  ['a', 1, 2.0]
    '''

    tree = NameToString().visit(ast.parse(string, mode='eval'))
    value = ast.literal_eval(tree)
    return value


def parse_known_args(parser, unparsed_args):
    '''Like ArgumentParser.parse_known_args but prints help on error.'''

    tmp_stdout = StringIO()
    tmp_stderr = StringIO()
    try:
        sys.stdout = tmp_stdout
        sys.stderr = tmp_stderr
        args, more_args = parser.parse_known_args(unparsed_args)
    except SystemExit:
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        if '-h' in unparsed_args or '--help' in unparsed_args:
            parser.print_help()
            sys.exit()
        print(tmp_stdout.getvalue())
        print(tmp_stderr.getvalue())
        raise
    finally:
        tmp_stdout.close()
        tmp_stderr.close()
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    return args, more_args


def run(cli, unparsed_args=missing):
    '''Execute a CLI command.'''

    if unparsed_args is missing:
        unparsed_args = sys.argv[1:]

    try:
        cli = cli()
    except TypeError as e:
        if 'object is not callable' not in str(e):
            raise

    # First parse used to pass arguments along to subcommands
    args, more_args = parse_known_args(cli.parser, unparsed_args)

    global_options = [('help', '-h'), ('verbose', '-v')]
    options = [f for n, f in global_options if args.__dict__.get(n, False)]

    if 'command' in args:
        cli_command = cli.get_command(args.command)
        if cli_command:
            # Recursively execute subcommand
            args = options + more_args
            return run(cli_command, args)

        if cli_command is not None:
            cli.parser.print_usage()
            print('error: unrecognized command "%s"' % args.command)
            print('       expected one of %s' % list(cli._commands.keys()))
            sys.exit(2)

    # Final parse for validation before run
    args = cli.parser.parse_args(unparsed_args)
    if args.help:
        cli.parser.print_help()
        sys.exit()

    return cli.run(args)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cpenv',
 'cpenv.repos',
 'cpenv.vendor',
 'cpenv.vendor.cachetools',
 'cpenv.vendor.certifi',
 'cpenv.vendor.shotgun_api3',
 'cpenv.vendor.shotgun_api3.lib',
 'cpenv.vendor.shotgun_api3.lib.httplib2',
 'cpenv.vendor.shotgun_api3.lib.httplib2.python2',
 'cpenv.vendor.shotgun_api3.lib.httplib2.python3',
 'cpenv.vendor.shotgun_api3.lib.mockgun',
 'cpenv.vendor.yaml',
 'cpenv.vendor.yaml.yaml2',
 'cpenv.vendor.yaml.yaml3']

package_data = \
{'': ['*'], 'cpenv': ['bin/*']}

install_requires = \
['colorama>=0.4.3,<0.5.0', 'psutil>=5.7.0,<6.0.0', 'tqdm>=4.46.0,<5.0.0']

entry_points = \
{'console_scripts': ['cpenv = cpenv.__main__:main']}

setup_kwargs = {
    'name': 'cpenv',
    'version': '0.5.14',
    'description': 'Cross-platform module and environment management.',
    'long_description': None,
    'author': 'Dan Bradham',
    'author_email': 'danielbradham@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)
